/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2000 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "bufed.h"
#include "c.h"
#include "grep.h"
#include "kill.h"
#include "tags.h"
#include "tex.h"
#include "info.h"
#include "diff.h"
#include "epsnet.h"
#include "filter.h"
#include "dired.h"
#include "makemode.h"
#include "html.h"
#include "perl.h"
#include "video.h"

// For updating from 10.00 or later.

when_loading()
{
	if (!strncmp(eel_version, "10b", 3)) {
		save_var bufnum = zap(HELPBUF);
		bprintf("%s is too old to be updated.\n"
				"The above state file is for Epsilon version %s.\n"
				"This Epsilon executable is from Epsilon version %s.\n",
				state_file, eel_version, EEL_MINOR_VERSION);
		view_help(" Update Not Performed ", 0);
		quick_abort();
	}
}

int inp_next_page();

normal_on_modify()
{
	int paging = (_buf_readonly == 3 || _buf_readonly && readonly_pages);

	if (_doing_input != DI_SEARCH)
		highlight_off();
	if (paging && key == ' ' && last_index == normal_character) {
		next_page();
		quick_abort();
	} else if (paging && (key == '\b' || key == GREYBACK)
			   && last_index == backward_delete_character) {
		previous_page();
		quick_abort();
	} else if (_buf_readonly)
		error("Buffer is read-only.");
	call_on_modify = 0;	/* don't call us next time */
	if (virtual_space && bufnum == window_bufnum)
		virtual_to_real_column();
}

command count_lines()
{
	int count, here, disksize, reg_lines;

	noteput("Counting...");
	here = lines_between(0, point, 1);
	check_abort();
	noteput("Counting... point on line %d...", here + 1);
	count = here + lines_between(point, size(), 1);
	check_abort();
	disksize = size();
	if (translation_type == FILETYPE_MSDOS)
		disksize += count;
	if (size() && character(size() - 1) != '\n')
		count++;
	save_var point, mark;
	fix_region();
	if (is_highlight_on()) {
		reg_lines = lines_between(point, mark, 1);
		if (mark > point && character(mark - 1) != '\n')
			reg_lines++;
		say("%d line%s, %d in region, point on line %d.  %d bytes on disk.",
			  count, count == 1 ? "" : "s", reg_lines, here + 1, disksize);
	} else
		say("%d line%s, point on line %d.  %d bytes on disk.",
			  count, count == 1 ? "" : "s", here + 1, disksize);
	expire_message = 1;
	iter = 0;
}

check_drop_buffer(bname) /* check if buffer can be deleted safely */
char *bname;		/* ask user if questionable */
{			/* return 1 if done, 0 if user said don't */
	char msg[80], *origbuf = bufname;
	int mod;

	if (!exist(bname))
		error("No buffer %s.", bname);
	if (another_process() && is_process_buffer(name_to_bufnum(bname)))
		error("A process is running in buffer %s.", bname);
	bufname = bname;
	mod = modified && size() > 0 && !discardable_buffer;
	if (ftp_job && ftp_job->buf == bufnum
		  && ftp_job->operation == FTP_SEND)
		mod = 1;
	bufname = origbuf;
	if (!another_process() && !strcmp(bname, PROCBUF))
		mod = 0;
	if (mod) {
		sprintf(msg, "Buffer %.40s not saved.  Delete it? ", bname);
		if (!ask_yn("Delete buffer", msg, "Delete", "Cancel", 2))
			return 0;
	}
	to_another_buffer(bname);
	return 1;
}

int move_before_func_name(char *func)
{
	int p, end;

	if (cind_try_space())
		return 0;
	point -= parse_string(-1, "</word>throws[ \t\n]+[a-z0-9_]+"
						  "([ \t\n]*,[ \t\n]*[a-z0-9_]+)*");
	if (cind_try_space())
		return 0;
	if (character(point - 1) != ')' || !c_rev_skip_level('(', ')'))
		return 0;
	if (cind_try_space())
		return 0;
	end = point;
	if (character(point - 1) == '>' && !parse_string(-1, "operator%>")
		  && (p = default_move_level(-1, ">", "<")) >= 0)
		point = p;			// Move past template stuff.
	if (!parse_string(-1, CPP_OPSPEC) && !parse_string(-1, "~?[a-zA-Z0-9_]+"))
		return 0;
	point = matchend;
	if (c_statement_start())
		return 0;
	if (func)
		if (end - point >= ptrlen(func) - 1)
			return 0;
		else
			grab(point, end, func);
	return 1;
}

int find_c_class_info(int start, char *type, char *class, char *func)
{
	char this[FNAMELEN], last[FNAMELEN];
	int class_start = point, i;

	point--;	// See if "class foo" is before this.
	strcpy(last, "");
	for (i = 20; i-- > 0; ) {	// Look through some words before the {.
		if (cind_try_space() || give_begin_line() > start)
			return 0;
		if (strchr("<>:,", character(point - 1)))
			point--;
		else {
			if (!parse_string(-1, "[a-zA-Z0-9_]+", this))
				return 0;
			if (!strcmp(this, "class") || !strcmp(this, "struct")
				  || !strcmp(this, "union") || !strcmp(this, "namespace")) {
				if (type)
					strcpy(type, this);		// Word to the right of type is class.
				if (class)
					strcpy(class, last);	// Is place where we started within func?
				if (find_inner_func_name(start, class_start, func))
					return CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC;
				if (func)
					*func = 0;
				return CF_INFO_TYPE | CF_INFO_CLASS;
			}
			strcpy(last, this);		// Go on to the next word.
		}
	}
	return 0;		// Couldn't recognize anything here.
}

int find_c_func_info(char *type, char *class, char *func, int stop_on_key)
{
	char funcname[FNAMELEN], classname[FNAMELEN];
	int start, funcstart;

	*classname = *funcname = 0;
	if (func)
		*func = 0;
	if (class)
		*class = 0;
	if (type)
		*type = 0;
	save_var point, matchstart, matchend, narrow_start = 0, narrow_end = 0;
	save_var tag_declarations = 0;	// We use tagging sometimes.
	to_begin_line();
	start = point;
	point--;		// Begin searching at the last } char on the left margin...
	while (search(-1, "\n}")) {
		point++;		// ...that's not in a comment.
		if (!in_c_comment(c_look_back))
			break;
		point--;
	}

	while (re_search(1, "{|/<*|/>|(~?[a-zA-Z0-9_]+|" CPP_OPSPEC ")[ \t\n]*%(")) {
		point = matchstart;
		if (stop_on_key && (char_avail() || is_key_repeating()))
			return -1;		// Give up search now?
		if (parse_string(-1, "^[ \t]*#[ \t]*define.*(\\\n.*)*")) {
			nl_forward();	// Ignore everything in a macro definition.
			continue;
		}
		switch (curchar()) {
			case '{':		// Could be start of class def.
				return find_c_class_info(start, type, class, func);
			case '/':
				if (character(point + 1) == '*') {		// Skip over comment.
					point += 2;
					search(1, "*/");
				} else
					nl_forward();	/* c++ comment */
				break;
			default:		// Found an identifier; is it a function def?
				if (give_begin_line() > start)
					return 0;		// Gone too far, so give up.
				funcstart = point;
				if (!parse_string(1, "~?[a-zA-Z0-9_]+|" CPP_OPSPEC, funcname))
					return 0;
				if (parse_string(1, "[ \t\n]*%("))
					narrow_end = size() - (point + 1000);
				if (tag_c_func(0) == 2) {	// Checks out as a function def.
					point = funcstart;
					*classname = 0;
					if (parse_string(-1, "[ \t\n]*::[ \t\n]*")) {
						point = matchend;
						parse_string(-1, "[a-zA-Z0-9_]+", classname);
					}
					if (func)
						strcpy(func, funcname);
					if (class)
						strcpy(class, classname);
					return CF_INFO_FUNC | (*classname ? CF_INFO_CLASS : 0);
				}
				*funcname = 0;
				narrow_end = 0;
				break;
		}
	}
	return 0;
}

find_c_func_name(char *funcname, int stop_on_key)
{
	char type[FNAMELEN], func[FNAMELEN], class[FNAMELEN];
	int i = find_c_func_info(type, class, func, stop_on_key);

	if (i == -1)		// User pressed a key, give up.
		return 0;
	*funcname = 0;
	if (strlen(type) + strlen(class) + strlen(func) + 4 >= ptrlen(funcname)) {
		sprintf(funcname, "...%s%.*s",
				(*class && *func) ? "::" : "",
				ptrlen(funcname) - 10, *func ? func : class);
		return 1;
	}
	switch (i) {			// Format the information for display.
		case CF_INFO_TYPE | CF_INFO_CLASS | CF_INFO_FUNC:
			sprintf(funcname, "%s %s::%s", type, class, func);
			break;
		case CF_INFO_TYPE | CF_INFO_CLASS:
			sprintf(funcname, "%s %s", type, class);
			break;
		case				CF_INFO_CLASS | CF_INFO_FUNC:
			sprintf(funcname, "%s::%s", class, func);
			break;
		case								CF_INFO_FUNC:
			sprintf(funcname, "%s", func);
			break;
	}
	return 1;
}

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	if (in_perl_buffer)
		return perl_compute_indent();
	norm_narrow = fix_narrowing(point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (!reindent_c_comments && parse_string(1, "[ \t\f]*/<*>"))
		return get_indentation(point);
	if (!reindent_one_line_c_comments && parse_string(1, "[ \t\f]*//"))
		return get_indentation(point);
	if (first_char == '#' || first_char == '\f')
		return 0;
	else if (first_char == '}')
		return cind_close();

	to_begin_line();
	if (character(point - 2) == '\\')
		return indent_after_backslash();
	if (parse_string(1, "[ \t]*\n") && after_many_comments()) {
		re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
		return get_indentation(point);
	}
	cind_space();			/* skip whitespace and comments */
	if (character(point - 1) == ':')
		move_before_labels();
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig)) {
		ind = indent_case(orig);
	} else if (parse_string(1, "ON_")) {
		// Found MFC dispatch table, leave alone.
	} else if (have_label(orig)) {
		char type[FNAMELEN];		// Could it be a bitfield, not a label?
		find_c_func_info(type, NULL, NULL, 0);
		if (!strcmp(type, "struct") || !strcmp(type, "union"))	// Yes, a bitfield.
			ind += get_c_indent();
		else					// It's a label.
			ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
		if (just_open(orig))
			ind += c_brace_offset;
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		if (just_open(orig) && func_start_block(pstart))
			ind = c_top_braces;
		else
			ind = indent_after_statement(orig);
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block(orig);
	} else if (prev_end == '{') {
		if (!c_indent_after_extern_c
			  && parse_string(1, "[ \t]*extern[ \t]*\"C\"[ \t\n]*{"))
			return ind;
		if (ind <= c_top_braces	// special case if no previous indent
			  || get_column(pend - 1) <= c_top_braces) {
			ind += top_indent(prev_end) ? get_c_indent() : c_top_braces;
			if (just_open(orig))
				ind += c_brace_offset;
			if (ind < c_top_braces)
				ind = c_top_braces;
		} else if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (just_open(orig))
		return indent_just_open(ind);
	else if (is_contin_statement(pend, 1, orig)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
			   && !(Closeback && just_open(point)))
		return ind;
	else
		ind += get_c_indent();
	return ind;
}

indent_after_block(orig)
{
	int ind;

	c_rev_skip_level('{', '}');	/* find matching { */
	to_begin_line();
	ind = get_indentation(point);
	if (Topindent && (ind == 0 || parse_string(1, "[ \t]*{")
					  && ind <= c_brace_offset))
		return 0;	// found a func def block
	return indent_after_statement(orig);
}

c_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], tmp[FNAMELEN], *p, *classpath, *include;
	int flags = PATH_ADD_CUR_DIR, is_abs = 0;

	to_begin_line();
	if (parse_string(1, "[ \t]*#[ \t]*include[ \t]*[<\"](.*)([>\"])"
					 "[ \t]*(/[*/].*)?$")) {
		grab(find_group(1, 1), find_group(1, 0), rel);
		is_abs = character(find_group(2, 1)) == '>'
				 && !find_linked_file_ignores_angles;
		// Using #include "" or <> syntax?
	} else if (is_java_buffer() &&  parse_string(1, "[ \t]*(package|import)"
		"[ \t]*(<word>+(%.<word>+)*(%.%*)?)[ \t]*;")) {
		grab(find_group(2, 1), find_group(2, 0), rel);
		while (p = strchr(rel, '.'))	// Transform java.lang to 
			*p = path_sep;				// java/lang filename.
		p = rel + strlen(rel) - 1;
		if (*p == '*') {
			strcpy(p, ".");
			flags |= PATH_PERMIT_DIRS;
		} else
			strcat(p, ".java");
	} else
		error("Can't find any link here.");
	strcpy(file, rel);
	get_buffer_directory(curdir);
	if (is_abs)				// Don't look for relative names at all.
		flags = 0;
	else {
		absolute(file, curdir);
		if (check_file(file))
			return;
	}
	flags |= PATH_PERMIT_WILDCARDS;
	include = getenv("INCLUDE");
	classpath = getenv("CLASSPATH");
	p = 0;
	if (!p && !strcmp(get_extension(filename), ".e")) {
		sprintf(tmp, "include%c%s", path_sep, rel);
		p = lookpath(tmp, 0);
	}
	if (!p && is_java_buffer() && classpath)
		p = look_on_path(rel, flags, classpath);
	if (!p && include)
		p = look_on_path(rel, flags, include);
	if (!p && *include_directories)
		p = look_on_path(rel, flags, include_directories);
	if (!p)
		p = look_on_path(rel, flags, (opsys == OS_UNIX)
						 ? INCL_PATH_UNIX : INCL_PATH_WIN);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

user short clipboard_format = CF_DEFAULT;

int should_use_clipboard()
{
	switch (clipboard_access) {
		case 0:	return 0;
		case 2:	return 1;		// Always, even in a macro.
	}
	return !in_macro() && !key_from_macro;
}

save_to_clipboard()	// try to save current kill buffer on clipboard
{
	int strip = clipboard_convert_newlines();

	if (should_use_clipboard() && clipboard_available()) {
		buffer_to_clipboard(name_to_bufnum(_cur_kill_buf), strip,
							clipboard_format);
		clipboard_to_buffer(zap("-clipboard-"), strip,
							clipboard_format);
	}
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	new = zap("-new-clipboard-");
	if (!clipboard_to_buffer(new, clipboard_convert_newlines(),
							 clipboard_format))
		return;
	old = create("-clipboard-");
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

when_restoring()
{
	int tflag = 0, i, j, splitmode = HORIZONTAL;
	char fname[FNAMELEN];
	char *s;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_DOS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	if (has_feature & FEAT_ANSI_CHARS) {
		word_pattern.default = default_word;
		_char_class.default = _def_ansi_char_class;
		_case_map.default = _def_ansi_case_map;
		_srch_case_map.default = _def_ansi_srch_case_map;
	} else {
		word_pattern.default = default_oem_word;
		_char_class.default = _def_char_class;
		_case_map.default = _def_case_map;
		_srch_case_map.default = _def_srch_case_map;
	}
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	early_init();

	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
			case 'f':
				switch (argv[i][2]) {
					case 's':
						set_swapname(argv[i] + 3);
						break;
					case 'h':
						set_shrinkname(argv[i] + 3);
						break;
					case 'd':
						strcpy(_docname, argv[i] + 3);
						break;
					default:
						bad_flag(argv[i]);
				}
				break;
			case 't':
				tflag = 1;
				break;
			case 'r':
			case 'l': break;
			case 'h': case '?': usage();
			case 'p':
				if (!argv[i][2])
					preserve_session = preserve_session ? 0 : 1;
				else if (isdigit(argv[i][2]))
					preserve_session = numtoi(argv[i] + 2);
				else
					cmd_line_session_file = argv[i] + 2;
				break;
			case 'v':
				switch (argv[i][2]) {
					case 'v':
						splitmode = VERTICAL;
						break;
					default:
						bad_flag(argv[i]);
				}
				break;
			case 'd':
				s = index(argv[i] + 2, '=');
				if (!s)		// accept either VAR=VAL or VAR!VAL
					s = index(argv[i] + 2, '!');
				if (s) {
					*s = 0;
					if (j = find_index(argv[i] + 2)) {
						set_var_val(j, s + 1, 1);
						break;
					}
					bprintf("No such variable %s.\n", argv[i] + 2);
					init_help_abort();
				}		/* fall through on error */
			default:
				bad_flag(argv[i]);
		}			/* i is 1st file argument */
	term_init();
#include "keyext.h"
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	if (tflag) {
		preserve_session = 0;
		tutorial();
	} else {
		if (load_from_state && (i >= argc || session_always_restore)) {
			if (session_restore_files == 2 && i < argc)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (i < argc)
			load_cmd_line_files(i, splitmode);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	set_toolbar();
	if (load_from_state && (strcmp(eel_version, versioned_file_string) < 0
							|| (strchr(eel_version, 'b')
								&& !strchr(versioned_file_string, 'b')))
		  && !strchr(versioned_file_string, 'a')) {
		// Each user has a private state file, so we can't update them
		// all automatically.
		save_var bufnum = zap(HELPBUF);
		bprintf("Your version %s state file requires updating.\n"
				"You need to start Epsilon with its -quickup flag.\n", 
				eel_version);
		bprintf("Until you do this, Epsilon will not work correctly.\n");
		view_help(" Update Needed! ", 0);
		restore_vars();
	}
	note("Epsilon Programmer's Editor %s, EEL version %s",
		 version, eel_version);
	try_calling("start-up");
	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
			case 'l': load_commands(argv[i] + 2); break;
			case 'r': try_calling(argv[i] + 2); break;
		}
}


usage()
{
	char name[FNAMELEN], *p;

	bufnum = zap("usage");
	bprintf("\n");
	strcpy(name, get_tail(argv[0], 0));
	*get_extension(name) = 0;
	if (opsys != OS_UNIX)
		for (p = name; *p; p++)
			*p = tolower(*p);
	bprintf("Usage: %s [flags...] [files...]\n", name);
	bprintf("A file name may have +linenum before it.\n\n");

	bprintf("Flags:\n");
	bprintf("-sFILE   Load state from FILE.sta (default: epsilon.sta).\n");
	bprintf("-bFILE   Build new system from FILE.b, not .sta file "
			"(default: epsilon.b).\n");
	bprintf("-quickup Load quickup.b after loading state."
			"    -nologo Skip initial msg.\n\n");

	bprintf("-lFILE   Load FILE.b after loading files. ");
	bprintf("-rCMD    Run command or macro CMD.\n");
	bprintf("-teach   Run the tutorial.                ");
	bprintf("-dVAR!VAL Set variable to value.\n");
	bprintf("-p[VAL]  Set/toggle session-preserving.   ");
	bprintf("-pFILE    Load session from FILE.\n\n");

	bprintf("-fsDIRS  Swap/shrink to directories DIRS. ");
	if (!(has_feature & FEAT_WINHELP))
		bprintf("-fdFILE  Documentation in file FILE.");
	bprintf("\n");
	bprintf("-fhDIRS  Shrink to directories in DIRS.\n\n");

	bprintf("-vmono   Use monochrome color scheme.     ");
	bprintf("-vlNUM   Screen has NUM lines.\n");
	bprintf("-vcolor  Use colors.                      ");
	bprintf("-vcNUM   Screen has NUM columns.\n");
	if (opsys == OS_DOS && !is_gui) {
		bprintf("-vclean  Fast video mode.                 ");
		bprintf("-vsnow   Slow video mode.\n");
		bprintf("-vmNUM   Screen memory is at NUM (hex).\n");
	}
	bprintf("-vv      Put cmd line files side by side.");

	if (!is_gui) {
		if (is_unix != IS_UNIX_XWIN)
			bprintf("\nPress any key for more flags.");
		file_write(TERMINAL, 1);
		leave_blank = 1;
		if (is_unix == IS_UNIX_TERM)
			term_position(0, screen_lines - 1);
		if (is_unix != IS_UNIX_XWIN)
			getkey();

		buf_zap(bufnum);
	}
	bprintf("\n\n");

	if (opsys == OS_DOS) {
		if (!is_gui && is_win32 != IS_WIN32_CONSOLE)
			bprintf("-kw      Run inside GUI window.           ");
		bprintf("-kaNUM   Kbd compatibility flags.\n");
	}
	if (opsys == OS_DOS && !is_gui || opsys == OS_OS2) {
		bprintf("-kcNUM   Mouse graphic cursor/block/off.  ");
		bprintf("-kmNUM   Mouse relative/absolute/off.\n");
	}
	if (opsys == OS_DOS && !is_gui && is_win32 != IS_WIN32_CONSOLE)
		bprintf("-kt      Don't take over system timer.    ");
	else if (opsys == OS_OS2)
		bprintf("-kpNUM   Wait NUM secs starting process.  ");
	bprintf("-ke      Extended key support.\n");
	if (opsys == OS_OS2) {
		bprintf("-ksNUM   Speed (>100) vs. abort response. ");
		bprintf("-ka      Kbd alternate mode.\n");
		bprintf("-kdNUM   Mouse delay in .001 secs [50].\n");
	}

	bprintf("\n");
	bprintf("-mNUM    Use NUM bytes of main memory.    ");
	bprintf("-m0      Use minimum memory.\n");
	bprintf("-mNUM    Use NUM kbytes (if NUM < 1000)   ");
	bprintf("-m       Use as needed.\n\n");
	if (opsys == OS_DOS && !is_gui && is_win32 != IS_WIN32_CONSOLE) {
		bprintf("-xNUM    Like -m; sets max XMS (extended) ");
		bprintf("-eNUM    Sets maximum EMS (expanded).\n");
		bprintf("-xiNUM   Initial allocation of XMS.       ");
		bprintf("-eiNUM   Initial allocation of EMS.\n");
		bprintf("-xfNUM   Allocation for EEL funcs in XMS. ");
		bprintf("-efNUM   Alloc. for EEL funcs in EMS.\n");
		bprintf("-xuNUM   Maximum use of upper memory blocks.\n\n");
	}
	bprintf("  NUMk means kilobytes, NUMm means megabytes.\n");
	bprintf("  Without suffix, NUM is in kilobytes if < 1000, "
			"otherwise bytes.\n");
	restore_vars();
	init_abort();
}

init_abort()	/* show current buffer, then abort */
{
	if (has_feature & FEAT_GUI_DIALOGS) {
		int buf = bufnum, win;
		window_bufnum = zap("Error");
		save_var _doing_input = DI_VIEW, paging_retains_view = 1;
		win = add_popup(0, 0, screen_cols, screen_lines, 0, buf);
		point = 0;
		go_to_link = 0;
		display_dialog_box("UsageBox", "Usage", win, 0, 0, "OK", "", "");
		view_loop(win);
		leave();
	} else {
		if (is_unix == IS_UNIX_TERM)
			bprintf("\nPress any key to continue.\n");
		file_write(TERMINAL, 1);
		leave_blank = 1;
		if (is_unix == IS_UNIX_TERM) {
			term_position(0, screen_lines - 1);
			getkey();
		}
		leave();
	}
}

when_idle_coloring(cnt)	// Called when idle, waiting for a key.
{			// Cnt increases while waiting for same key.
	int last = 0, from, to;

	if (idle_coloring_delay < 0 || !(coloring_flags & COLOR_DO_COLORING)
		  || (coloring_flags & COLOR_IN_PROGRESS) || !size())
		return -1;
	if (cnt < 1)
		time_begin(&idle_color_time, idle_coloring_delay);
	if (!time_done(&idle_color_time))
		return time_remaining(&idle_color_time);
	save_var narrow_start = 0, narrow_end = 0;
	if (!get_tagged_region("needs-color", 0, &from, &to)) {
		if (to >= size())	// We're done if the good region extends to eof.
			return -1;		// Otherwise, get the following (bad) region.
		get_tagged_region("needs-color", to, &from, &to);
	}
	last = from;
	save_var point, matchstart, matchend, abort_searching = 0;
	coloring_flags |= COLOR_IN_PROGRESS; // ensure errors don't restart us
	color_for_area(last, last + idle_coloring_size);
	coloring_flags &= ~COLOR_IN_PROGRESS;	// no errors aborted us, do more
	return 0;
}

color_c_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	char pat[200];

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	talk = (to - from > 20000);	// show status during long delays
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
			   "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*");
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '/':			// found // one-line comment
				while (nl_forward() && character(point - 2) == '\\')
					;		// Continued // comment.
				set_character_color(t, point, color_class c_comment);
				break;
			case '*':			// found /* starting comment
				search(1, "*/");
				set_character_color(t, point, color_class c_comment);
				break;
			case '#':			// found preproc line
				c_preproc_color();	// move past correct part
				set_character_color(t, point,
									color_class c_preprocessor);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, c_keyword_color(t));
				if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
					  && (point - t == 4)			// uuid's in IDL files.
					  && parse_string(-1, "uuid")
					  && parse_string(1, "%([0-9A-Fa-f]"))
					maybe_color_uuid();
				break;
		}
		if (talk && point > talk_at + 2000) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk)
		note("");
	return point;
}

char *f_only_match();

set_menu_look(int win)
{
	display_more_msg(win);
	text_color = (is_gui && want_gui_prompts) ? color_class dialog_gui_text
											  : color_class menu;
	mouse_cursor = &left_pointer;
}

pass_back(cmd)	/* send cmd to caller to be processed */
{
	cmd_result = cmd;
	if (_doing_input == DI_VIEWLAST)
		remove_final_view();
	else if (_doing_input == DI_LISTBOX)
		return;
	else if (_doing_input)
		leave_recursion(cmd);
	else
		do_command(binding_of(key, root_keys));
}

pass_if_complete(cmd)	/* if we're doing completion, pass it, else */
{			/* run default definition of key */
	if (_doing_completion)
		pass_back(cmd);
	else
		do_command(binding_of(key, root_keys));
}

inp_get_and_exit()
{
	if (insert_newlines_in_dialog)
		do_command(binding_of(key, root_keys));
	else
		pass_back(INP_GET_EXIT);
}

inp_next_page()
{
	if (_doing_input) {
		if ((all_blanks(window_end, size() - narrow_end) + next_page())
			  && !paging_retains_view)
			if (_doing_input == DI_VIEWLAST)
				remove_final_view();
			else
				leave_recursion(INP_SEEN);
	} else
		do_command(binding_of(key, root_keys));
}

inp_previous_page()
{
	if (_doing_input) {
		if (search_in_menu)
			pass_back(INP_ADD);
		else if (previous_page() && !paging_retains_view)
			if (_doing_input == DI_VIEWLAST)
				remove_final_view();
			else
				leave_recursion(INP_SEEN);
	} else
		do_command(binding_of(key, root_keys));
}

complete(s, finder, flags)	/* perform completion by modifying s */
char *s;		/* return 0: no matches, 1: 1 match, */
char *(*finder)();	/* else many matches, or some matches */
{				/* don't start with s */
	char biggest[FNAMELEN], *t;
	int i, start, end, num = 1, c;

	if (!(t = (*finder)(s, STARTMATCH | flags)))
		return 0;				/* if no matches */
	strcpy(biggest, t);			/* start off with first */
	start = (flags & CAUTIOUS) ? 0 : strlen(s);
			/* assume all matches start with s unless cautious */
	save_var case_fold = (flags & COMP_FOLD) != 0;
	end = strlen(biggest);
	while (t = (*finder)(s, flags)) {
		if (strnfcmp(t, biggest, end))	/* if this differs, */
			for (i = start; i < end; i++) /* find where */
				if (charfcmp(t[i], biggest[i]))
					end = i;
		num++;
		if (end <= start)	/* if all chars gone, stop */
			break;
	}
	if ((flags & COMP_FILE) == COMP_FILE)
		// it's a file; use special folding
		for (i = 0; s[i]; i++) { // compare with std path chars
			c = (s[i] != ':' && is_path_separator(s[i])) ? path_sep
				: _srch_case_map[s[i]];
			if (end < i || c != _srch_case_map[biggest[i]])
				return 2; // some matches didn't start with s
		}
	else if (flags & CAUTIOUS) {
		if (strnfcmp(s, biggest, strlen(s)) || end < strlen(s))
			return 2; // some matches didn't start with s
	}
	strncpy(s, biggest, end);
	return num;
}

restart_file_paths()
{
	char pat[FNAMELEN];
	int res, res2;

	if (insert_default_response != 3)
		return;
	save_var point = 0, abort_searching = 0, case_fold = 1;
	point += parse_string(1, "(ftp|http|telnet):[\\/][\\/][^\\/]+");
		// Don't consider ftp://host.com part, which may have \\ after.
	if (opsys == OS_UNIX) {
		res = parse_string(1, "/.*/!~");
		if (res)	// Delete /dir/name/ before a ~.
			delete(0, point + res);
	}
	sprintf(pat, "((ftp|http|telnet%s):)?"	// optional a: or ftp:
			"[\\/].*[\\/]!"			// slash, stuff, slash
			"([\\/]|(ftp|http|telnet|[a-z]):)",		// slash or a: or ftp:
			(opsys == OS_UNIX) ? "" : "|[a-z]");
	res = parse_string(1, pat);
	if (res) {	// Delete abc:\dirname\dir2\ or \dir\ or k:\dir\, etc.
		point += res;	// But exclude a trailing URL part.
		res2 = parse_string(-1, "(ftp|http|telnet):[\\/]+");
		if (res2 && res2 < point)
			point -= res2;
		delete(0, point);
	}
	if (opsys != OS_UNIX) {
		res = parse_string(1, "[a-z]:[\\/]!([\\/]|[a-z]+:)");
		if (res)	// Delete c:\ or similar before an absolute path.
			delete(0, point + res);
	}
}

view_linked_buf(buf, last, linker)
int (*linker)();
{
	int win, res, height, prev;

	remove_final_view();
	prev = window_handle;
	if (!last)
		save_var _doing_input = DI_VIEW, bufnum, window_handle;
	height = screen_lines - _view_top - _view_bottom;
	win = add_popup(_view_left, _view_top,
					screen_cols - _view_left - _view_right, height,
					_view_border, buf);
	go_to_link = linker;
	if (linker) {
		create_links();
		mouse_dbl_selects = 1;
	}
	if (is_gui && want_gui_prompts)
		if (!linker)
			one_window_to_dialog(_view_title ? _view_title : "No Title", win,
								 "OK", "Cancel", "");
		else
			one_window_to_dialog(_view_title ? _view_title : "No Title", win,
								 "", "Done", "");

	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	if (_view_title)
		window_title(win, BTOP, TITLECENTER, _view_title);
	point = 0;
	window_to_fit(height);
	if (last) {
		_view_prev_input = _doing_input;
		_view_prev_window = prev;
		_view_prev_mode = mode_keys;
		mode_keys = view_tab;		/* switch buf to view mode */
		_doing_input = DI_VIEWLAST;
		_window_flags &= ~FORCE_MODE_LINE;
		display_more_msg(win);
	} else {
		res = view_loop(win);
		if (res == INP_EXIT && !IS_ENTER_KEY(key))
			ungot_key = key;
		remove_window(win);
	}
	user_abort = 0;
}

#define MAXARG 200000000

command argument() on reg_tab[CTRL('U')], reg_tab[ALT('-')],
					  reg_tab[NUMALT(GREYMINUS)]
{
	int seendigit, fixkey, newiter, first = 1, sign = 1;

	iter = 1;
	seendigit = 0;
	for (;;) {
		fixkey = key & 0xff; /* strip meta bit: use A-1, A-2, etc */
		if (IS_EXT_ASCII_KEY(key) && isdigit(fixkey)) {
			newiter = (seendigit ? (10 * iter) : 0) + fixkey-'0';
			seendigit = 1;
		} else if (IS_EXT_ASCII_KEY(key) && fixkey == '-') {
			newiter = first ? 4 : iter;
			sign = -sign;
		} else if (first || binding_of(key, root_keys)
				   == find_index("argument")) {
			newiter = iter * 4;
			seendigit = 0;
		} else if (key != MOUSE_MOVE)
			break;
		first = 0;
		if (newiter <= MAXARG)
			iter = newiter;
		mention("Argument: %d ", iter * sign);
		getkey();
	}
	iter *= sign;
	has_arg = 1;
	say("");

	do_topkey();
	has_arg = 0;
	iter = 1;
}

int binding_of(int k, short *tab)
{
	return (k & CMD_INDEX_KEY) ? (k & ~CMD_INDEX_KEY) : tab[k];
}

char _fundamental_mode_name[] = "Fundamental";

mode_default_settings()
{
	indenter = indenter.default;
	if (!indenter)
		indenter = prev_indenter;
	auto_indent = auto_indent.default;
	buffer_maybe_break_line = 0;
	strcpy(comment_start, comment_start.default);
	strcpy(comment_pattern, comment_pattern.default);
	strcpy(comment_begin, comment_begin.default);
	strcpy(comment_end, comment_end.default);
	mode_keys = buf_tab;		/* use these keys */
	major_mode = _fundamental_mode_name;
	when_displaying = 0;
	recolor_range = recolor_from_here = 0;
	drop_coloring(bufnum);
	make_mode();
}

zeroed TIMER idle_display_time;
int auto_show_delimiter_delay;	// time before next check (.01 sec units)

when_idle(cnt)
{
	int delay, disp;

	if (!cnt || time_done(&idle_display_time)) {
		show_idle_text();
		time_begin(&idle_display_time, 10);	// wait .1 seconds next time
	}
	if ((auto_show_matching_characters || mode_auto_show_delimiter) && cnt == 1) {
		auto_show_delimiter_delay = 5;
		if (maybe_auto_show_delimiter() == 2)	// No match, wait .5 sec
			auto_show_delimiter_delay = 50;	 // before looking on next key.
		refresh();
	}
	if (!leave_blank && cnt == 2 && set_display_func_name())
		maybe_refresh();
	delay = when_idle_coloring(cnt);
	if (*show_when_idle) {		// Wait for .1 seconds.
		disp = time_remaining(&idle_display_time);
		if (delay < 0 || disp < delay)
			delay = disp;
	}
	if (cnt < 2 && (delay < 0 || auto_show_delimiter_delay < delay))
		delay = auto_show_delimiter_delay;
	return (cnt < 2) ? 0 : delay;
}

do_dired_examine(fname, isdir)
char *fname;
{
	char file[FNAMELEN], *p;

	save_var file_pattern_wildcards = 0;	// []'s etc in file name are literal.
	strcpy(file, fname);
	absolute(file);
	if (isdir == 3 && !is_remote_file(file))
		isdir = 0;	// symbolic link to a local file; get directly
	if (isdir) {
		if (is_remote_file(file) && isdir == 1
			  && is_path_separator(file[strlen(file) - 1]) != 1)
			strcat(file, "/");	// If dir, not link, add /.
		if (dired_one(file) || is_remote_file(file))
			return;
		strcpy(file, fname);
		p = get_tail(file, 0);	/* for .., go to right place */
		if (p > file && strcmp(p, "..") == 0) {
			p[-(opsys != OS_UNIX)] = 0;	// Remove / too unless Unix.
			p = get_tail(file, 0);
			strcat(p, "\n");
			col_search(p, goal_column);
			move_to_column(goal_column);
			build_first = 1;
		}
	} else			/* regular file */
		find_it(file, ask_line_translate());
}

dired_sort()
{
	char *p, msg[200];
	int rev, sort;

	if (dired_format == DF_VMS)
		error("Sorting VMS directories not supported.");
	rev = isupper(dired_sorts_files);
	switch (sort = toupper(dired_sorts_files)) {
		case 'N': case 'D': case 'S':  case 'E': 
			break;
		case 0: case 'U':
			sort = 'U', rev = 0; break;
		default:
			sort = 'N';
	}
	for (;;) {
		strcpy(msg, "Name, Ext, Size, Date, Unsorted,");
		p = index(msg, sort);
		p = index(p, ',');
		msg[strlen(msg) - 1] = 0;
		sayput("Sort by %.*s (now)%s; Reverse; %sGroup dirs"
			   " [NESDURG] ", p - msg, msg, p,
			   dired_groups_dirs ? "un" : "");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'N':
			case 'D':
			case 'S':
			case 'E': sort = toupper(key); break;
			case 'U': say("Future listings will be unsorted.");
					  sort = 0; break;
			case 'R': rev = !rev; break;
			case '+': rev = 0; break;
			case '-': rev = 1; break;
			case 'G': dired_groups_dirs = !dired_groups_dirs;
					  break;
			default:  if (IS_ENTER_KEY(key))
						  break;
					  else if (binding_of(key, root_keys) == (short) help
							   || key == 'h' || key == '?')
						  help_on_command("dired-sort");
					  continue;
		}
		break;
	}
	dired_sorts_files = rev ? toupper(sort) : tolower(sort);
	if (sort) {
		sprintf(msg, "%s%s, %s", rev ? "Reverse " : "",
				(sort == 'N') ? "Name" : (sort == 'E') ? "Extension"
			: (sort == 'S') ? "Size" : "Date",
				dired_groups_dirs ? "grouping directories at top"
								  : "not grouping directories");
		noteput("Sorting...");
		sort_for_dired();
		say("Sorted by %s.", msg);
	} else
		say("Future listings will be unsorted.");
	maybe_refresh();
}

dired_do_delete(name, to, in_dired)	/* return 1 if error */
char *name, *to;
{
	int isdir = in_dired ? dired_isdir() : 0;

	to = to;
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER,
					  isdir ? "RMD %s\n" : "DELE %s\n",
					  get_url_file_part(name, 0));
	} else {
		if (in_dired)
			noteput("Deleting %s.", name);
		if (isdir ? rmdir(name) : delete_file(name))
			return 1;
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

dired_print()
{
	char fname[FNAMELEN];
	int had, old = window_bufnum, err = 0, repeated = has_arg, newbuf;

	save_var has_arg = 0, iter = 1;
	dired_getname(fname);
	had = look_file(fname);
	if (had)
		to_buffer(bufname);
	else
		err = find_in_other_buf(fname, ask_line_translate());
	newbuf = bufnum;
	if (!err)
		err = do_print_region(bufnum, 1, repeated);
	to_buffer_num(old);
	if (!had)
		buf_delete(newbuf);
	if (err)
		quick_abort();
}

command quick_dired_command()
{
	char file[FNAMELEN], rel[FNAMELEN], msg[FNAMELEN], dest[FNAMELEN];
	char dir[FNAMELEN];

	strcpy(file, filename);
	strcpy(dir, filename);
	*get_tail(dir, 0) = 0;
	if (!*file)
		error("This buffer has no associated file.");
	strcpy(rel, get_tail(file, 0));
again:
	do {
		sayput("%s: D=Delete, Move, Copy, View, T=Props, .=Dired, "
			   "G=Go dir, ?=Show more: ", rel);
		getkey();
		check_abort();
	} while (!IS_TRUE_KEY(key));
	say("");
	if (is_remote_file(file) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	switch (toupper(key)) {
		case '.':
			dired_one(file);
			break;
		case 'D':
			sprintf(msg, "Delete %s? ", rel);
			if (!ask_yn("Delete file", msg, "Delete", "Cancel", 1))
				break;
			if (dired_do_delete(file, NULL, 0))
				file_error(errno, file, "Could not delete");
			else if (!is_remote_file(file))
				say("%s deleted.", file);
			break;
		case 'C':
			if (is_remote_file(file))
				error("Copy operation not implemented for remote files.");
			sprintf(msg, "Copy %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				switch (dired_do_copy(file, dest, 0)) {
					case 1:
						file_error(errno, dest, "Could not copy");
						break;
					case 0:
						say("%s copied.", file);
						break;
				}
			}
			break;
		case 'M':
			sprintf(msg, "Move %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				if (dired_do_move(file, dest, 0))
					file_error(errno, dest, "Could not move");
				else if (!is_remote_file(file))
					say("%s moved.", file);
			}
			break;
		case 'V':
			do_shell_execute(file, "Open", "run viewer");
			break;
		case 'T':
			do_shell_execute(file,  "Properties", "show properties");
			break;
		case 'F':
			do_shell_execute(dir, "Open", "run viewer");
			break;
		case 'G':
			if (chdir(dir))
				file_error(errno, dir, "unknown error");
			else {
				getcd(dir);
				say("Current directory: %s", dir);
			}
			break;
		case '+':
			do_create_directory(0);
			break;
		case '!':
			do_shell_command(file, dir);
			break;
		default:
			if (key == '?' || key == HELPKEY || key == FKEY(1))
				help_on_command("quick-dired-command");
			else
				say("Invalid key.");
			goto again;
	}
	if (is_remote_file(file))
		send_remote_ftp_op(file);
}

command dired()
{
	char dir[FNAMELEN];

	tiled_only();
	get_file_dir(dir, "Edit directory: ");
	if (was_quoted)		// []'s etc in file name are literal.
		save_var file_pattern_wildcards = 0;
	return dired_one(dir);
}

detect_dired_format()
{
	int old = dired_format;

	dired_format = 0;
	goal_column = NAME_START;
	save_var point = 0, case_fold = 1;
	while (parse_string(1, SKIPLINEPAT))
		nl_forward();
	if (point >= size())
		return;
	if (parse_string(1, ".*" UNIX_DATE)) {
		dired_format = DF_UNIX;
		goal_column = get_column(matchend);
	} else if (parse_string(1, ".*" OLDNT_DATE)) {
		dired_format = DF_OLDNT;
		goal_column = 41;
	} else if (parse_string(1, ".*" VMS_DATE)) {
		dired_format = DF_VMS;
		goal_column = 2;
	} else if (!parse_string(1, ". [ 0-9][0-9]-[a-z?][a-z?][a-z?]-[0-9][0-9]")) {
		dired_format = DF_SIMPLE;
		goal_column = 2;
	}
	if (dired_format != old)
		drop_all_colored_regions();
}

command set_tab_size()
{
	tab_size = get_number("Tab size: ");
	if (tab_size < 2 || tab_size > 256)
		tab_size = 8;
	say("Tab size is %d in this buffer.", tab_size);
	when_setting_tab_size();
	build_first = 1;
}

when_setting_tab_size()
{
	char varname[FNAMELEN], msg[FNAMELEN];
	int i;

	sprintf(varname, "%s-tab-override", major_mode);
	i = find_index(varname);
	sprintf(msg, "Set the mode-specific tab size variable %s to %d also?",
			varname, tab_size.default);
	if (i && get_num_var(i) != tab_size.default && get_num_var(i) > 0)
		if (ask_yn("Set override too?", msg, "Yes", "No", 1))
			set_num_var(i, tab_size.default);
}

set_display_func_name()
{
	char subr[80];
	int i;

	if (!display_definition || is_window(window_handle) != ISTILED)
		return 0;
	display_func_name_buf = bufnum;	// Only show it in this buffer
	display_func_name_win = window_handle;	// or window.
	save_var permit_window_keys |= PERMIT_WHEEL_KEY;
	sprintf(subr, "%s_func_name_finder", major_mode);
	i = find_index(subr);
	switch (name_type(i)) {		// i has name table index of subr/var
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(int (*)())i)();		// Cast i to a func ptr, then call it.
			break;
		default:					// No func names for this buffer.
			*display_func_name = 0;
	}
	if (display_func_name_method() == 1)
		make_mode();
	return 1;
}

// If m contains a %f sequence, replace it by the function name if any,
// doubling any %'s within.

substitute_func_name(char *m)
{
	char *p, *s, tmp[FNAMELEN];

	for (p = m; p = strchr(p, '%'); ) {	// look for %f in string.
		if (p[1] == 'f') {
			strcpy(tmp, p + 2);
			if (display_func_name_method() == 1 && *display_func_name) {
				for (s = display_func_name; *s && ptrlen(s) > 1
						 && ptrlen(p) > 2; ) {
					if (*s == '%')
						*p++ = '%';
					*p++ = *s++;
				}
				*p++ = ' ';
			}
			if (ptrlen(p) > strlen(tmp) + 2)
				strcpy(p, tmp);
		} else		// Skip past %% sequences.
			p += p[1] ? 2 : 1;
	}
}

check_dates(save) /* compare this file's time info w/ version on disk */
{		/* 0 means ok, 1 means cancel save */
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
		  || check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);
	if ((auto_read_changed_file || _force_auto_read_changed_file)
		  && !save && maybe_auto_read(&now))
		return 1;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 22;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			char junk[FNAMELEN];
			view_buf(temp, 0);
			if (user_abort)
				break;
			getkey();
			if (key == WIN_DRAG_DROP)	// Ignore dropped files.
				while(drag_drop_result(junk))
					;
		}
		key = toupper(key);
		if (key == 'C') {
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, translation_type);
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, translation_type))
				say("%s: Read from disk.", filename);
			point = pos;
			to_begin_line();
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 1;	// Don't check on getting focus.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

command copy_to_file()
{
	int mod = modified;
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;
	char tmp[FNAMELEN], orig[FNAMELEN];

	iter = 0;
	strcpy(orig, filename);
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Copy to file", &flags, 1))
			return;
	} else
		get_file(tmp, "Copy to file",
				 force_save_as == FSA_NEWFILE ? "" : filename);
	if (warn_existing_file(tmp))
		return;
	filename = tmp;
	save_var already_made_backup = 0, force_save_as = 0;
	do_save_file(1, 0, 0);
	filename = orig;
	modified = mod;
}

command visit_file()
{
	char tmp[FNAMELEN];

	iter = 0;
	save_var show_mouse_choices = 1;
	get_file(tmp, "Visit file", force_save_as == FSA_NEWFILE ? "" : filename);
	if (!*tmp)
		return;
	if (!has_arg && modified) {	/* buffer need to be saved? */
		if (ask_save_buffer())
			do_save_file(1, 1, 1);
	}
	read_file(tmp, translation_type);
}

command revert_file()
{
	char pr[FNAMELEN], rel[FNAMELEN];

	save_var point;
	if (!*filename || force_save_as)
		error("No file for this buffer.");
	iter = 0;
	if (modified) {		/* buffer need to be saved? */
		relative(filename, rel);
		sprintf(pr, "Discard changes and revert to %s? ", rel);
		if (!ask_yn("Discard changes", pr, "Discard", "Cancel", 1))
			return;
	}
	read_file(filename, translation_type);
	build_first = 1;
	fix_window_start();
}

command insert_file()
{
	char file[FNAMELEN];
	int flags = OFN_HIDEREADONLY;

	save_var show_mouse_choices = 1;
	if (use_common_file_dialog()) {
		strcpy(file, _region_file);
		if (!use_common_open_file_dlg(file, "Insert file", &flags, 0))
			return;
	} else {
		get_buffer_directory(_region_file);
		get_file(file, "Insert file", _region_file);
	}
	if (!*file)
		return;
	strcpy(_region_file, file);
	mark = point;			/* put mark before it */
	set_region_type(REGNORM);
	build_first = 1;
	do_insert_file(file, ask_line_translate());
	iter = 0;
}

convert_cygwin_filename(char *file)
{
	char buf[FNAMELEN];

	if (opsys != OS_DOS || !cygwin_filenames || file[0] != '/'
		  || file[1] != '/' || !isalpha(file[2]) || file[3] != '/')
		return;
	strcpy(buf, file + 3);		// Rewrite //f/file into f:/file.
	sprintf(file, "%c:%s", file[2], buf);
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")) {
		sprintf(res, pat, "mak");	// Recognize special file name
		return 1;					// w/o .mak extension.
	}
	if (mode_keys == telnet_tab) {
		sprintf(res, pat, "telnet");
		return 1;
	}		
	save_var point = 0;
	if (parse_string(1, "#%!.*/perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(sh|csh|bash)</word>|:\n")) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		  && re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (translation_type == FILETYPE_UNIX && parse_string(1, "([ \t]*\n)*#")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

wild_pattern(s)		/* does this filename use the {},; wildcards? */
char *s;
{
	return is_pattern(s) & 1;
}

region_fill_first_line(int a, int b, int first_width, int rest_width)
{
	int start, atend, len;

	if (a > b)			/* switch so a < b */
		start = a, a = b, b = start;
	save_spot point = a;
	save_var narrow_start = a, narrow_end = size() - b;
	save_var abort_searching = 0;
	point = 0;		/* remove previous line breaks */
	while (re_search(1, "( [ \t\n]+)|([\t\n][ \t\n]*)")) {
		len = point - matchstart;
		atend = character(matchstart) == '\n';
		delete(matchstart, point);
		if ((len >= 2 || atend) && parse_string(-1, "[.?!][])'\"]*",
												NULL))
			stuff("  ");
		else
			insert(' ');
	}
	point = 0;
	widened_move_to_column(first_width);
	for (;;) {		/* put in new line breaks */
		if (point >= size() - narrow_end)
			break;
		re_search(-1, " +");
		if (!re_search(1, " +"))
			break;
		delete(matchstart, point);
		insert('\n');
		narrow_start = point;	/* never look before here */
		widened_move_to_column(rest_width);
	}
	build_first = 1;		/* redisplay hint */
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], search_type[80], *(*finder)();
	spot start;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	getcd(abs);
	bprintf("Found 0 lines with %s\"%s\" in %s (cur dir \"%s\")\n",
			search_type, pat, filepat, abs);
	set_character_color(before_head, point, color_class grep_header);
	start = alloc_spot(0);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning |= ROWARN_GREP;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (!inbufs && !ok_grep_file_match(s)) {
			skipped++;
			continue;
		}
		grep_make_names(rel, abs, s, inbufs);
		noteput("%d matches, searching %s...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			check_dates(0);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		res = copy_matches(name_to_bufnum(buf), flags, pat, rel);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			point = before_head;
			delete(point, point + parse_string(1, "Found [0-9]+ "));
			bprintf("Found %d ", cnt);
			set_character_color(before_head, point, color_class grep_header);
			point = size();
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		modified = 0;
		point = i;
		if (skipped)
			say("Skipped %d files due to grep-ignore-file-extensions", skipped);
	}
}

prompt_for_help()
{
	for (;;) {
		mention("Help on (? for choices): ");
		do {
			getkey();
		} while (!IS_TRUE_KEY(key));
		check_abort();
		say("");
		if (binding_of(key, root_keys) == (short) help || key == '?')
			dodesc("help", "help", 0);
		else if (key < 128)
			switch (tolower(key)) {
				case 'k': describe_key(); return;
				case 'c': describe_command(); return;
				case 'a': apropos(); return;
				case 'b': show_bindings(); return;
				case 'q': what_is(); return;
				case 'l': show_last_keys(); return;
				case 'v': about_epsilon(); return;
				case 'r': describe_variable(); return;
				case 'm': help_on_mode(); return;
				case 'i': info(); return;
				case 'f': epsilon_info_look_up(); return;
				case 'w': windows_only_help();
						  win_help_contents(""); return;
				case 'h': epsilon_html_look_up(); return;
				case CTRL('C'): info_goto_epsilon_command(); return;
				case CTRL('K'): info_goto_epsilon_key(); return;
				case CTRL('V'):
				case CTRL('R'): info_goto_epsilon_variable(); return;
			}
	}
}

make_man_page(char *topic)
{
	char cmdline[FNAMELEN], buf[FNAMELEN];
	int color;

	sprintf(cmdline, "man %s", topic);
	sprintf(buf, "man-%s", topic);
	quiet_set_bookmark();
	to_buffer(buf);
	man_mode();
	if (!size()) {
		set_read_only(0);
		pipe_text(NULL, buf, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		point = 0;
		if (parse_string(1, "Reformatting.*please wait.*\n"))
			delete(0, matchend);
		simple_re_replace(1, "\n\n\n+", "\n\n");
		set_character_color(0, size(), -1);
		point = 0;
		while (re_search(1, ".\b.")) {
			if (character(point - 3) == '_')
				color = color_class man_underlined;
			else
				color = color_class man_bold;
			delete(point - 3, point - 1);
			set_character_color(point - 1, point, color);
			point--;
		}
		filename = "fake-man-file.man";	// So find-linked-files works.
		force_save_as = FSA_NEWFILE;
		discardable_buffer = buffer_not_saveable = 1;
	}
	modified = point = 0;
	set_read_only(3);
}

html_recolor_from_here(safe)
{
	int start = point;

	safe = safe;			// avoid compiler warning
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!"))
		return point;
	point = start;
	if (re_search(-1, "%<%!--|-[ \t\n]*>|%<[%?]|[%?]%>"))
		return point;
	return start;
}

color_html_attributes(script)
{
	int len, start, java = 0;

	while (len = parse_string(1, "[-_./a-z0-9]*")) {	// An attribute?
		if (script && parse_string(1, "(language|type)[ \t\n]*=[ \t\n]*"
								   "[\"']*(text/)?(javascript|jscript)"))
			java = 1;
		set_character_color(point, point + len, color_class html_attribute);
		point += len;			// Followed by a value?
		if (len = parse_string(1, "[ \t\n]*=[ \t\n]*")) {
			set_character_color(point, point + len, color_class html_punctuation);
							// Color the = and whitespace.
			point += len;		// Value might be quoted.
			start = point;
			if (curchar() == '"') {
				point++;
				search(1, "\"");
			} else if (curchar() == '\'') {
				point++;
				search(1, "'");
			} else	// No quotes, assume it's whitespace-delimited.
				re_search(1, "[^ \t\n<>]*");
			set_character_color(start, point, color_class html_value);
		}
		re_search(1, "[ \t\n]*");	// Parse whitespace.
		set_character_color(matchstart, matchend, color_class html_punctuation);
	}
	if (curchar() == '>') {
		set_character_color(point, point + 1, color_class html_punctuation);
		point++;
	}
	start = point;
	if (script) {
		if (search(1, "</script>"))
			point = matchstart;
		color_embedded_java(start, point, !java || !html_javascript_coloring);
	}
}

color_embedded_java(from, to, plain)
{
	if (plain) {
		set_character_color(from, to, color_class html_attribute);
		return;
	}
	save_var point;
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	save_var c_extra_keywords |= JAVA_KEYWORDS;
	save_var recolor_range = color_c_range;
	save_var recolor_from_here = color_c_from_here;
	color_for_narrowed_area(from, to);
}

color_html_comment()
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (parse_string(1, "%<[ \t\n]*%![ \t\n]*-"))
		// A legal comment?
		re_search(1, "-[ \t\n]*>");	// Search for end of legal comment.
	else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

color_html_include()
{
	int s, c;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	set_character_color(point, point + 2, color_class html_attribute);
	s = point;
	search(1, c == '%' ? "%>" : c == '?' ? "?>" : ">");
	set_character_color(point - 2, point, color_class html_attribute);
	color_embedded_java(s + 2, matchstart, !html_javascript_coloring);
}

color_html_move_back()
{
	int start = point, p;

	save_var narrow_start = point - color_html_look_back;
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!"))
		return;
	point = start;
	if (re_search(-1, "<Langle>[ \t\n]*/?script")
		  && parse_string(1, "<Langle>[ \t\n]*script"))
		return;				// Found a <script>, not a </script>.
	point = start;
	if (re_search(-1, "<Langle>[%?]|[%?]<Rangle>")
		  && curchar() == '<')
		return;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!search(-1, "<"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			search(-1, "<!");	// go back to its start.
		}
	} while (parse_string(-1, "[\"'].*"));
}

get_soft_tab_size()
{
	return (soft_tab_size > 0) ? soft_tab_size : tab_size;
}

indent_like_tab()	/* indent like tab char, but maybe use spaces */
{
	int cur = current_column(), tab = get_soft_tab_size();

	to_column(cur + tab - cur % tab);
}

maybe_indent_rigidly(int rev)
{
	if (!is_highlight_on())
		return 0;
	if (rev) {
		if (!has_arg)
			iter = -get_soft_tab_size();
		else if (iter > 0)
			iter = -iter;
		has_arg = 1;
	}
	indent_rigidly();
	highlight_on();
	shift_selecting = 2;
	return 1;
}

command back_to_tab_stop()
{			/* move back to nearest tab stop */
	int tab = get_soft_tab_size(), old;

	if (maybe_indent_rigidly(1))
		return;
	old = current_column();
	move_to_column(((current_column() - 1) / tab) * tab);
	if (old && old == current_column())
		point--;
}

command indent_rigidly()
{
	int temp, indent;

	if (!has_arg) iter = get_soft_tab_size();
	fix_region();
	if (nl_reverse())	/* to start of line */
		point++;
	temp = point;
	while (point < mark) {
		re_search(1, "[ \t]*");
		indent = current_column() + iter;
		to_column(indent);
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
	}
	mark = temp;
	iter = 0;
}

show_menu_loop(num, cmd)
char *cmd;
{
	int sub = 0, line = 0, max, isdir, oldsub, orignum;

	max = max_submenus();
	if (IS_MOUSE_KEY(key))
		ungot_key = key;
	do {
		oldsub = sub;
		sub = to_sub_line(sub, num, line);
		if (getkey() == MOUSE_MOVE) {
			refresh();
			while (getkey() == MOUSE_MOVE)
				;
		}
		isdir = 1;		// assume it's a direction key
		switch (get_direction()) {
			case BTOP:	line--; break;
			case BBOTTOM:	line++; break;
			case BLEFT:	num--, line = 0; break;
			case BRIGHT:	num++, line = 0; break;
			default:	isdir = 0; break;
		}
		if (isalpha(key))
			line = do_menu_search(key);
		else if (IS_MOUSE_DOWN(key)) {
			if (oldsub && mouse_y <= 0)	// maybe drop menu
				orignum = count_submenus(mouse_x) - 1;
			else
				orignum = -2;	// upclick on bar won't drop
			sub = while_mouse_pressed(sub);
			num = after_mouse_pressed(sub, num, orignum);
			line = 0;
		} else if (binding_of(key, reg_tab) == (short) help) {
			get_menu_cmd(cmd, sub);
			if (*cmd && find_index(cmd))
				help_on_command(cmd);
			isdir = 1;
		}
		num = (num + max) % max;
		line = (line + window_height) % window_height;
	} while (isdir || isalpha(key) || IS_MOUSE_KEY(key));
	if (!exit_menu_doit() && !exit_menu_cancel()) {
		ungot_key = key;
		key = ESC;
	}
	return sub;
}

// Ignore the wheel key; it's already been handled.
mouse_wheel_key() on reg_tab[WIN_WHEEL_KEY], cx_tab[WIN_WHEEL_KEY]
{
}

drag_drop_handler()
{
	char buf[FNAMELEN], *s, dir[FNAMELEN];
	int line = -1, column = -1, j, dir_flags = directory_flags, first = 1;

	tiled_only();
	if (mouse_screen >= 0)
		to_mouse_window(mouse_x, mouse_y, mouse_screen);
	while(drag_drop_result(buf)){
		if (*buf == '+') {			// a +linenum parameter
			line = numtoi(buf + 1);
			s = strchr(buf, ':');
			if (s)
				column = numtoi(s + 1);
		} else if (*buf == '-') {
			switch (buf[1]) {
				case 'd':
					s = index(buf + 2, '=');
					if (!s)		// accept either VAR=VAL or VAR!VAL
						s = index(buf + 2, '!');
					if (!s)
						break;
					*s = 0;
					if (j = find_index(buf + 2)) {
						set_var_val(j, s + 1, 1);
						continue;
					} else
						say("No such variable %s.", buf + 2);
					break;
				case 'l':
					load_commands(buf + 2);
					break;
				case 'r':
					try_calling(buf + 2);
					break;
				case 'w':
					dir_flags = numtoi(buf + 2);
					break;
			}
			continue;
		} else if (*buf) {
			absolute(buf);
			if (is_directory(buf) || is_remote_file(buf))
				do_find(buf, default_translation_type);
			else
				find_it(buf, default_translation_type);
			if (line >= 0) {
				go_line(line);
				line = -1;
			}
			if (column >= 0) {
				move_to_column(column);
				column = -1;
			}
			if ((dir_flags & DIR_CWD_FROM_FILE) && first) {
				strcpy(dir, buf);
				*get_tail(dir, 1) = 0;
				chdir(dir);
			}
			first = 0;
		}
	}
}

no_running(int talk)			/* make sure no process is running */
{
	if (!is_process_buffer(name_to_bufnum(PROCBUF)))
		return 0;
	if (talk)
		say("A process is already running in buffer %s.", PROCBUF);
	if (strcmp(bufname, PROCBUF)) {
		tiled_only();
		locate_window(PROCBUF, "");
		maybe_refresh();
	}
	point = size();
	return 1;
}

command start_process()
{
	char cmd[FNAMELEN], *cmdline, *sh, fixup[FNAMELEN], *p;
	char cust1[FNAMELEN], cust2[FNAMELEN], buf[FNAMELEN];
	int i = 1;

	_last_was_grep = 0;
	tiled_only();
	error_if_input(1);
	strcpy(buf, PROCBUF);
	if (has_arg && (has_feature & FEAT_MULT_CONCUR)) {
		if (is_process_buffer(name_to_bufnum(buf)))
			do {
				sprintf(buf, "process-%d", i++);
			} while (is_process_buffer(name_to_bufnum(buf)));
	} else if (no_running(1))
		return;
	iter = has_arg = 0;
	cmdline = get_cmdline("Start Process With Command", cmd);
	create_concur(cmdline, buf);
	locate_window(buf, "");
	if (opsys == OS_UNIX && (sh = getenv("SHELL"))) {
		*cust1 = *cust2 = *fixup = 0;
		if (p = lookpath("epsilon.sh", 0))
			strcpy(cust1, p);
		if (p = lookpath("epsilon.csh", 0))
			strcpy(cust2, p);
		if (!*cust1 || !*cust2)
			sprintf(fixup, "echo \"Customization file not found, "
					"no customizations performed.\">/dev/null");
		else if (strstr(sh, "/sh") || strstr(sh, "/bash") || strstr(sh, "/ksh"))
			sprintf(fixup, ". %s; set +o emacs", cust1);
		else if (strstr(sh, "/csh") || strstr(sh, "/tcsh"))
			sprintf(fixup, "source %s", cust2);
		else
			sprintf(fixup, "echo \"Unrecognized shell, no customizations "
					"performed.\">/dev/null");
		delay(40, 0);	// Let's first get a prompt if we can.
		point = type_point;
		bprintf("%s\n", fixup);
	}
	point = size();
	process_mode();
}

create_concur(char *cmdline, char *buf)
{
	char dir[FNAMELEN], old[FNAMELEN];
	int err = 0;

	create(buf);
	if (clear_process_buffer)
		zap(buf);
	if (start_process_in_buffer_directory) {
		getcd(old);	// Don't switch to proc buffer before
		get_buffer_directory(dir);	// doing this.
		chdir(dir);
	}
	bufname = buf;
	point = size();
	if (!is_process_buffer(bufnum))
		err = do_concur_shell("", cmdline, buf);
	if (start_process_in_buffer_directory == 2)
		chdir(old);
	if (err) {
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
}

do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (sep_buf_ok) {
		bufnum = create(procbuf = "compilation-output");
		tab_size = process_tab_size;
		discardable_buffer = 1;
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (!concur_shell("", cmd, curdir, procbuf)) {
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE) {
					ungot_key = key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY);	// Delay 0.1 secs
			}
			err = process_exit_status;
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else
		err = do_push(cmd, 1, 0);
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

maybe_restart_concurrent()
{
	if (!(has_feature & FEAT_MULT_CONCUR) && restart_concurrent
		  && !another_process())
		do_concur_shell("", "", PROCBUF);
}

int concur_handler();

do_concur_shell(char *prog, char *cmd, char *buf)
{
	save_var bufnum = create(buf);
	when_activity = concur_handler;
	restore_vars();
	return concur_shell(prog, cmd, "", buf);
}

// A reference to this file was found in compiler output in the
// specified buffer.  Determine the directory it's in and make it an
// absolute pathname.  We recognize the "Running" message we insert,
// and the Entering and Leaving messages created by Gnu Make that
// indicate it's changing directories.

process_find_directory(char *file, int buf, int pos)
{
	char dir[FNAMELEN];
	int level = 0, got = 0;

	save_var bufnum = buf, point = pos;
	*dir = 0;
	if (is_process_buffer(bufnum) && *process_current_directory)
		strcpy(dir, process_current_directory);
	while (!got && re_search(-1, "(Entering|Leaving|Running `.*' in)"
							 " directory `(.*)'$")) {
		switch (character(find_group(1, 1))) {
			case 'L':	level++; break;
			case 'E':	if (--level > 0)
							break;
			case 'R':	got = 1;
						grab(find_group(2, 1), find_group(2, 0), dir);
						break;
		}
	}
	convert_cygwin_filename(dir);
	if (*dir)
		absolute(file, dir);
	else
		absolute(file);
}

/* if you change ERROR_PATTERN, change the next 3 macros as well */
#define ERROR_PATTERN	"^(e \"|w \"|<*>*(error|warning|fatal|[a-z0-9]+ )" \
			"( [a-z][0-9]+)?<*>*:?|[\"(]?) *" \
			"(([a-z]:)?[+-9A-z~][+-9A-z~ ]*<.>[+-9A-z]+[0-9<>]*)\"?" \
			"([ (:,]( *line|L)?) *" \
			"([0-9]+)(<rparen|:>|[^0-9\n].*[:;]).*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */
#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

#define ERROR_PATTERN2	".* at (.+) line ([0-9]+)%."
#define FILE_PAREN2	1	/* filename in () pair above */
#define LINE_PAREN2	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN3	"^([+-9A-z~:]+[0-9<>]*)([ (:,]( line|L)?) *" \
						"([0-9]+)[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN3	1	/* filename is in 1st () pair above */
#define LINE_PAREN3	2	/* line# is in 2nd pair of parens above */

// You can define ERROR_PATTERN4 with an alternate pattern.  Epsilon
// will look for either one, on lines that contain SIMP_PATTERN.
// Define FILE_PAREN4 and LINE_PAREN4 to match.  (Also ERROR_PATTERN5
// etc.)

do_next_error(int count, char *cmd)	// return 0 if no errors, else 1
{
	char file[FNAMELEN], number[30], line[130], *p;
	int lineno, origpt, dir, from, to, is_bufname, file_paren, line_paren;
	int orig = bufnum, procbuf;

	tiled_only();
	_last_was_grep = 0;
	if (spot_to_buffer(error_spot) <= 0 && !exist(PROCBUF))
		error("No process buffer to read errors from.");
	tiled_only();
	if (spot_to_buffer(error_spot) > 0)
		bufnum = spot_to_buffer(error_spot);
	else
		bufname = PROCBUF;
	procbuf = bufnum;
	origpt = point;
	case_fold = 1;
	if (spot_to_buffer(error_spot) > 0)
		point = *error_spot;
	else {
		error_spot = alloc_spot();
		point = 0;
	}

	noteput("Searching...");
	if (!count)
		find_error(-1, &file_paren, &line_paren), count = 1;
	dir = (count > 0) ? 1 : -1;
	for (; count; count -= dir) {
		if (!find_error(dir, &file_paren, &line_paren)) {
			point = origpt;
			bufnum = orig;
			if (_had_errors)
				say("No more errors.");
			else if (cmd)
				say("No errors running \"%s\".", cmd);
			else
				say("No errors.");
			return 0;
		}
	}
	_had_errors = 1;
	grab(find_group(file_paren, 1), find_group(file_paren, 0), file);
	grab(find_group(line_paren, 1), find_group(line_paren, 0), number);
	from = find_group(line_paren, 0);
	to = find_group(0, 0);
	if (to - from < 5)	// error message at end is too short
		from = find_group(0, 1);	// so try including whole line
	if (to > from + sizeof(line) - 1)
		to = from + sizeof(line) - 1;
	grab(from, to, line);
	is_bufname = (parse_string(1, "Buffer ") != 0);
	if (dir > 0)
		to_end_line();
	*error_spot = point;
	point = origpt;
	lineno = strtoi(number, 10);
	bufnum = orig;
	if (is_bufname)
		locate_window(file, "");
	else {
		process_find_directory(file, procbuf, from);
		locate_window("", file);
		find_it(file, default_translation_type);
	}
	go_line(lineno);
	for (p = line; *p && index("() :;\t", *p); p++)
		;			// skip over these chars at start of msg
	say("=> %s", p);
	return 1;
}

/* Find next error in direction, return 1 and set vars to find_group() paren count */
find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

do_when_process_output(int buf, int start, int end)
{
	save_var bufnum = buf;
	set_character_color(start, end, color_class process_output);
	if (opsys == OS_UNIX)
		color_ansi_output(start);
/*	if (window_bufnum == buf && point == size()
		&& end > window_end && process_output_to_window_bottom)
		line_up_to_bottom();*/
	save_var position_window_on_screen_line = 100;
	refresh();
}

#define LEFTD	"[({"
#define RIGHTD	"])}"

init_auto_show_delimiter()
{
	if (!show_matching_start_end_spot) {	// Initialize.
		show_matching_start_spot = alloc_spot(1);
		show_matching_end_spot = alloc_spot(1);
		show_matching_start_end_spot = alloc_spot(1);
		show_matching_end_end_spot = alloc_spot(1);
		if (modify_region(SHOW_MATCHING_REGION, MRTYPE, -1) != -1)
			remove_region(SHOW_MATCHING_REGION);
		add_region(show_matching_start_spot, show_matching_start_end_spot,
				   color_class show_matching_start, 0,
				   SHOW_MATCHING_REGION);
		add_region(show_matching_end_spot, show_matching_end_end_spot,
				   color_class show_matching_end, 0,
				   SHOW_MATCHING_REGION);
		modify_region(SHOW_MATCHING_REGION, MRCONTROL,
					  find_index("is_current_window"));
	}
}

try_auto_show_delimiter(int dir, char *near, char *far)
{
	int orig, there, res;

	save_var point;
	if (dir > 0) {	// Only go one line forward, and not too far back.
		save_var narrow_start = give_begin_line();
		save_var narrow_end = size() - (point + c_look_back);
	} else {		// Only go one line back, and not too far forward.
		save_var narrow_start = ++point - c_look_back;
		save_var narrow_end = size() - give_end_line();
	}
	orig = point;
	res = move_level(dir, near, far, 0, 1);
	if (res != 1)
		return 0;
	there = point;
	res = move_level(-dir, far, near, 0, 1);
	if (res != 1 || point != orig)
		return 0;
	*show_matching_end_spot = there - (there > point);
	*show_matching_end_end_spot = *show_matching_end_spot;
	*show_matching_start_spot = point - (there < point);
	*show_matching_start_end_spot = *show_matching_start_spot;
	return 1;
}

hide_auto_show_delimiter()
{
	if (show_matching_start_end_spot)	// Remove existing highlight.
		modify_region(SHOW_MATCHING_REGION, MRTYPE, 0);
}

int maybe_auto_show_delimiter()
{
	char *p, i, rightside, adjacent = 0;

	if (show_matching_start_end_spot) {	// Remove existing highlight.
		if (spot_to_buffer(show_matching_start_spot) != bufnum) {
			show_matching_start_spot = 0;	// A spot in some other buffer.
			show_matching_start_end_spot = 0; // (we must have copied it).
			show_matching_end_spot = 0;
			show_matching_end_end_spot = 0;
		}
		modify_region(SHOW_MATCHING_REGION, MRTYPE, 0);
	}
	if (num_of_popups() > 0 || is_key_repeating())
		return 0;
	if (mode_auto_show_delimiter) {
		i = (*mode_auto_show_delimiter)();
		if (i != 1)
			return i;
	}
	if (!auto_show_matching_characters)
		return 0;
	p = index(auto_show_matching_characters, curchar());
	if (!p && auto_show_adjacent_delimiter) {
		adjacent = 1;	// Check prev character, maybe it's a delim.
		save_var point--;
		p = index(auto_show_matching_characters, curchar());
	}
	if (!p)
		return 1;				// Not a delimiter we can match.
	i = p - auto_show_matching_characters;		// Index of char.
	rightside = i * 2 >= strlen(auto_show_matching_characters);
	if (adjacent && !(auto_show_adjacent_delimiter & (rightside ? 1 : 2)))
		return 1;	// Maybe reject some adjacent delimiters.
	save_var point, matchstart, matchend, abort_searching = 0;
	init_auto_show_delimiter();
	if (rightside ? try_auto_show_delimiter(-1, RIGHTD, LEFTD)
				  : try_auto_show_delimiter(1, LEFTD, RIGHTD)) {
		modify_region(SHOW_MATCHING_REGION, MRTYPE, REGINCL);
		return 3;			// Show the two regions.
	}
	return 2;
}

safe_copy_buffer_variables(to, from)
{
	copy_buffer_variables(to, from);
//	save_var bufnum = to;
}


	/* s_code: where to start searching from */
#define FHERE	1	/* from right here */
#define FORIG	2	/* from starting position */
#define FMATCH	3	/* from last match */
#define FMADD	4	/* same, but indicates we added a char this time */
#define FMREGEX	5	/* same, but don't assume too much about pattern */

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, mstart, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, okpoint, orig, ostart;
	int oreg_type, old_mark = -1;
	char tmpbuf[FNAMELEN];

	if (_doing_input == DI_SEARCH)
		error_if_input(0);
	if (is_highlight_on() && point != mark && ABS(point - mark) < 200
		  && !*trystr && typing_deletes_highlight
		  && !in_macro() && _len_def_mac < 1) {
		old_mark = mark;	// Use this as the initial search string.
		highlight_off();
	}
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = DI_SEARCH;
	save_var last_index;
	save_var mark = point, _highlight_control;
	save_var postpone_asynchronous_notify = 1;
	save_var macro_runs_immediately = 0;	// Make special keys work in macros.
	oreg_type = region_type();
	set_region_type(REGNORM);
	flags &= ~SEARCH_EMPTY;
	orig = okpoint = mstart = point;
	ostart = window_start;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
					 pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	trystr = malloc(FNAMELEN);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
						0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	if (old_mark != -1) { // Start out with highlighted region.
		window_handle = origwin;
		buf_xfer(editbuf, old_mark, point);
		window_handle = editwin;
		point = 0;
		if (flags & REGEX)	// Quote special chars.
			string_replace("[[%.!#()*+|?^$<]", "%#0", REGEX);
		point = size();
		highlight_on();		// Mark it so user can delete by typing.
	}

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
							   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key))
			if (mouse_y == echo_line)
				flags &= ~INCR;
			else {
				flags |= SONCE;
				ungot_key = key;
				break;
			}
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			point = okpoint;
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
				lastok = size();
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:
				stuff(default_search_string(flags));
				lastok = size();
			} else if (oldflags & FAILING) {
				if (search_wraps) {
					flags |= (flags & SEARCH_WRAPPED) ? SEARCH_OVERWRAPPED
						: SEARCH_WRAPPED;
					window_handle = origwin; /* Wrap to other end of buffer. */
					point = (flags & REVERSE) ? size() : 0;
					window_handle = editwin;
					s_code = FHERE;			/* And search again. */
				} else
					flags |= FAILING; /* no more matches, stop */
			} else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if ((flags & INCR) && size())
				s_code = FMATCH; /* if modes changed, search again */
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (binding_of(key, reg_tab) == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (binding_of(key, reg_tab) == (short) backward_delete_character
				   && (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
										((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (binding_of(key, comp_tab) == (short) inp_recall
				   && (IS_ALT_KEY(key) || !(flags & INCR))) {
			flags = pick_prev_search(flags) & ~INCR;
			if (user_abort)
				break;
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if ((binding_of(key, comp_tab) == (short) inp_grab_word
				 || binding_of(key, comp_tab) == (short) inp_grab_line)
				  && (IS_ALT_KEY(key) /*|| !(flags & INCR) */ )) {
				if (!(oldflags & FAILING)) {
					if (grab_search_word(flags, origwin,
										 (binding_of(key, comp_tab) == (short) inp_grab_line))) {
						window_handle = origwin;	// Move origin
						mstart = orig = point;		// to start of word.
						window_handle = editwin;
						s_code = FORIG;
					} else
						s_code = (flags & REVERSE) ? FHERE : FMREGEX;
				}
			} else if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		_doing_input = DI_SEARCH;
		if (!is_window(origwin) || !is_window(editwin)
			  || !is_window(pwin)) {	// user deleted our window
			user_abort = 1;
			goto end;
		}
		user_abort = 0;
		grab_expanding(0, size(), &trystr, 256);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			point = orig;
		else if (s_code >= FMATCH) {
			point = mstart;
			if (flags & REVERSE)
				point++;
		}
		_highlight_control = 0;
		switch (do_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
			case 0:	flags |= FAILING;
					point = okpoint;
					if (in_macro())
						goto end;
					else if (!(oldflags & FAILING))
						maybe_ding(bell_on_search);
					break;
			case 1:	okpoint = point;
					mark = mstart = matchstart;
					highlight_on();
					if ((flags & SEARCH_WRAPPED) && (point > orig) == !(flags & REVERSE))
						flags |= SEARCH_OVERWRAPPED;
					window_handle = editwin;
					goodchars = lastok = size();
					break;
			case DSABORT:
				point = okpoint;
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				break;
		}
	}
	window_handle = origwin;
	if (key == abort_key || user_abort)
		point = orig, window_start = ostart;
	else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
							  : &_default_flags) = flags;
			grab_expanding(0, size(), &trystr, 256);
			if (_last_was_regex)
				grab_expanding(0, size(), &_default_regex_search, 256);
			else
				grab_expanding(0, size(), &_default_search, 256);
			save_search_recall(flags, trystr);
		} else
			flags |= SEARCH_EMPTY;
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ESC)
			ungot_key = key;
	}
end:
	strncpy(tmpbuf, trystr, FNAMELEN - 1);
	free(trystr);
	if (is_window(origwin))
		window_handle = origwin;
	set_region_type(oreg_type);
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if ((flags & FAILING) && in_macro())
		error("%s: not found", tmpbuf);
	check_abort();
	say("");
	restore_vars();
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

exit_incr()	/* does this key exit an incremental search? */
{
	return key != '\t' && key != CTRL('Q') && (key < ' ' || !IS_ASCII_KEY(key))
			&& ((binding_of(key, comp_tab) != (short) inp_grab_word
				 && binding_of(key, comp_tab) != (short) inp_grab_line)
				|| !IS_ALT_KEY(key)) && !norm_binding();
}

norm_binding()	/* does this key run normal-character? */
{
	return ((key & CMD_INDEX_KEY) ? (key & ~CMD_INDEX_KEY)
								  : (mode_keys[key] ? mode_keys[key] : root_keys[key]))
			== (short) normal_character;
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, null_detect = -1, new_null_detect;
	int perform_case_replace, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= 0) : (point++ >= size())) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
				/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
							show = 1;
							goto ask;
						}
						if (!IS_ENTER_KEY(key))
							ungot_key = key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = check_for_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
				case ',': case '^':	skip = 1; goto ask;
				case '.':		ret = 1; goto done;
				case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro())
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
				  replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

command write_state()
{
	char fname[FNAMELEN], *p;

	build_filename(fname, "%x%f", default_state_file_name);
	if (opsys == OS_UNIX && *state_file && (strcmp(state_file, fname)
											|| get_file_read_only(state_file) == 1)) {
				// Suggest first directory of EPSPATH.
		if (p = verenv("EPSPATH")) {
			strcpy(state_file, p);
			p = strchr(state_file, path_list_char);
			if (p)
				*p = 0;
		} else		// Or if none, off home dir.
			strcpy(state_file, "~/.epsilon");
		absolute(state_file);		// Strip off any trailing /.
		sprintf(state_file + strlen(state_file), "/%s",
				default_state_file_name);
	} else if (!*state_file)
		strcpy(state_file, default_state_file_name);
	save_var only_file_extensions = "|.sta|";
	get_file(fname, "Write current state on file", state_file);
	if (!*fname)
		return;

	do_save_state(fname);
	strcpy(state_file, fname);
	iter = 0;
}

do_save_state(fname)		/* save state in file, aborting on error */
char *fname;
{
	int err;
	char file[FNAMELEN], backupname[FNAMELEN];

	strcpy(file, fname);
	copy_bindings();	/* fix up changed bindings */
	sayput("Writing...");

	strcpy(get_extension(file), state_extension);
	if (want_state_file_backups) {
		build_filename(backupname, state_file_backup_name, file);
		make_backup(file, backupname);
	}
	if (err = save_state(file)) {
		file_error(err, file, "write error");
		maybe_ding(bell_on_write_error);
		quick_abort();
	} else
		say("State saved on %s.", file);
}

command create_file_associations()
{
	int buf = zap("File Associations");
	char ext[FNAMELEN], cmd[FNAMELEN], dir[FNAMELEN], tmp[FNAMELEN];

	save_var window_bufnum;
	to_buffer_num(buf);
	stuff("# Associate Epsilon with the following extensions:\n");
	stuff("# Delete or add lines as required, then type Ctrl-X Ctrl-Z\n");
	stuff("# (the exit-level command) to proceed.  This command cannot\n");
	stuff("# remove existing Epsilon associations, only add new ones.\n");
	stuff("\n.c\n.cpp\n.h\n.e\n.txt\n");
	conf_mode();
	recursive_edit();
	if (strcmp(bufname, "File Associations"))
		error("Canceled.");
	point = 0;
	make_registration_entry("Epsilon", "Epsilon");
	get_executable_directory(dir);
	strcpy(tmp, argv[0]);
	absolute(tmp, dir);
	if (is_gui == IS_WIN31)
		sprintf(cmd, "%s %%1", tmp);
	else
		sprintf(cmd, "\"%s\" \"%%1\"", tmp);
	make_registration_entry("Epsilon\\shell\\open\\command", cmd);
	make_registration_entry("Epsilon\\shell\\open\\ddeexec", "\"%1\"");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\application",
							"Epsilon");
	make_registration_entry("Epsilon\\shell\\open\\ddeexec\\topic",
							"Open");
	while (point < size() - narrow_end) {
		grab_line(bufnum, ext);
		if (*ext == '.')
			make_registration_entry(ext, "Epsilon");
		else if (*ext && *ext != '#')
			say("Ignoring invalid extension \"%s\"", ext);
	}
	restore_vars();
	buf_delete(buf);
	say("File associations created.");
}

zeroed char tags_from_bsc;

int tag_lister();

tag_searcher_c(char *tag)
{
	char classname[FNAMELEN], fulltag[FNAMELEN], pr[FNAMELEN];
	int err, cnt = 0, selected = 0, try = 1;

	if (parse_string(-1, "-%>|%.[ \t\n]"))	// Ignore func's class.
		*classname = 0;
	else
		find_c_func_info(NULL, classname, NULL, 0);
again:
	strcpy(fulltag, tag);
	if (!strchr(tag, ':')) {		// Class not known yet.
		if (*classname && strcmp(classname, tag)) {		// See if current class 
			sprintf(fulltag, "%s::%s", classname, tag);	// has such a member.
			if (!(err = go_tag(fulltag, 1)))
				return err;
			bufnum = window_bufnum;
		}			// Is there more than one function by that name in any class?
		*fulltag = 0;
		bufnum = window_bufnum;
		if (tag_match(tag, STARTMATCH | EXACTONLY)) {
			do {
				cnt++;
			} while (tag_match(tag, EXACTONLY));
			if (cnt != 1) {
				sprintf(pr, "%d matches for %s: ", cnt, tag);
				save_var menu_width = 60, completion_lister = tag_lister;
				save_var insert_default_response = 2, tag_list_exact_only = 1;
				ungot_key = '?';
				comp_read(fulltag, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
						  | MUST_MATCH | NONE_OK | CAUTIOUS, tag);
				selected = 1;
			} else
				strcpy(fulltag, tag_match(tag, STARTMATCH | EXACTONLY));
		} else
			strcpy(fulltag, tag);
	}
	err = go_tag(fulltag, 1);
	strcpy(_last_tag, tag);
	if (err && !selected) {		// Try retagging current file before giving up.
		bufnum = window_bufnum;
		sprintf(pr, "Tag %s has moved, retag file %s? ",
				fulltag, filename);
		if (!tag_ask_before_retagging || try > 1
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			if (try < 2)
				do_retag_file(filename);
			err = go_tag(fulltag, 1);
			if (try < 2)
				do_save_tags();
			if (err && try++ < 2)
				goto again;
			if (!err)
				say("Found %s in this file.", tag);
		}
	}
	if (!err && !selected && strcmp(tag, fulltag))
		say("Searched for %s, found %s", tag, fulltag);
	report_tag_error(err);
}

tag_c_struct(int flavor)
{
	char func[TAGLEN];
	int start;

	c_fwd_space();
	start = point;
	if (!parse_string(1, "[a-zA-Z0-9_]+", func)) {
		if (flavor == 2 || flavor == 3)
			*func = 0;	// namespace or enum doesn't require a name
		else
			return 1;
	}
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (flavor == 1 && !is_java_buffer() && parse_string(1, "[a-zA-Z0-9_]+")) {
		parse_string(1, "[a-zA-Z0-9_]+", func);
		c_fwd_space();
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		point++;	// skip over { so we do stuff within
		strcpy(c_tagging_class, func);	// remember, now in class
	} else if (flavor == 2)
		tag_c_enum();
	if (*func)
		add_tag(func, start);
	return 1;
}

list_definitions_when_displaying(from, to)
{
	char tagname[FNAMELEN], msg[FNAMELEN];

	from = from;		// not used
	if (1) {
		set_wattrib(window_handle, PVERTBORDCOLOR, color_class text);
		set_wattrib(window_handle, PHORIZBORDCOLOR, color_class text);
//		set_wattrib(window_handle, PTITLECOLOR, color_class text);
		text_color = color_class text;
	}
	to = tag_line_to_position(lines_between(0, point, 0) + 1, tagname);
	if (*tagname) {
		sprintf(msg, " %s - %q ", tag_declarations ? "Select a declaration or definition"
			: "Select a definition", tagname);
		window_title(window_handle, BTOP, TITLELEFT(2), msg);
		set_window_caption(window_handle, msg);
	}
	save_var window_handle = _bufed_orig_window;
	point = to;
	position_to_tag_line();
}

command list_definitions()
{
	char tag[FNAMELEN], pat[FNAMELEN], *title;
	int tags = zap("-list-definitions"), tmp = tmp_buf(), pos, decls = 0;
	int orig = bufnum, oldpoint = point, starthere = -1, old_ws = window_start;
	int orig_tab_size;

	delete_buffer("-tag-errors");
	iter = 0;
	save_var point = 0, _tags_buffer = tags, tag_show_percent = 1;
	save_var tag_by_text = 1;
	if (has_arg)
		save_var tag_declarations = decls = 3;
	if (!*filename || !call_by_suffix(filename, "tag-suffix-%s"))
		error("Don't know how to list definitions in this buffer.");
	orig_tab_size = tab_size;
	if (!buffer_size("-list-definitions")) {
		save_var tag_declarations = decls = 3;
		point = 0;
		call_by_suffix(filename, "tag-suffix-%s");
	}
	if (!buffer_size("-list-definitions"))
		error("No definitions or declarations found in this buffer.");
	say("");
	restore_vars();
	if (want_code_coloring && recolor_range)
		recolor_buffer_range(0, size());	// Update colors for whole buffer.
	save_var point, bufnum = tags;
	point = 0;		// Copy colored lines of original file into tmp, based on tags.
	while (point < size() && parse_string(1, "^[^\t\n]+\t[^\t\n]+\t([^\t\n]+)\t")) {
		grab(find_group(1, 1), find_group(1, 0), pat);
		pos = strtoi(pat, 10);		// Position in original buffer.
		bufnum = orig;
		point = pos;
		if (give_begin_line() > oldpoint && starthere < 0)	// Look for starting pos.
			starthere = buffer_size(bufnum_to_name(tmp)) - 1;
		buf_xfer_colors(tmp, give_begin_line(), give_end_line() + 1);
		bufnum = tags;
		nl_forward();
	}
	bufnum = tmp;
	delete_final_newlines();
	point = (starthere == -1) ? size() : starthere;
	to_begin_line();	// Begin at corresponding point to original location.
	goal_column = 0;
	tab_size = orig_tab_size;
	when_displaying = list_definitions_when_displaying;
	restore_vars();
	*tag = 0;
	save_var menu_width = screen_cols - _view_left - _view_right;
	save_var _bufed_orig_window = window_handle;	// Save it here.
	save_var tag_declarations = decls;
	title = tag_declarations ? "Select a declaration or definition"
							 : "Select a definition";
	if (get_choice(tmp, tag, title, title, "OK", "Cancel", "")) {
		save_var bufnum = tmp;
		pos = tag_line_to_position(lines_between(0, point, 0) + 1, NULL);
		restore_vars();
		point = oldpoint;
		quiet_set_bookmark();
		point = pos;
		position_to_tag_line();
	} else
		point = oldpoint, window_start = old_ws;
	buf_delete(tmp);
	buf_delete(tags);
}

command next_video()
{				/* pick some different video mode */
	char name[30], *s, i = 0, len;

	if (is_gui)
		error("Resize a window by dragging its border.");
	else if (opsys == OS_UNIX || is_win32)
		error("Can't change video modes in this environment.");
	iter = 0;
	if (*current_video_mode)
		strcpy(name, current_video_mode);
	else
		sprintf(name, "%dx%d", screen_cols, screen_lines);
	len = strlen(name);
	ati_maybe_blank(1);
	s = vid_match("", 1);		// find current mode in list
	while (s && (strncmp(name, s, len) || (current_video_mode ? s[len]
		: isdigit(s[len])))
		   && (s = vid_match("", 0)))
		check_abort();
	do {
		if (!(s = vid_match("", 0)))	/* get next one, or first */
			if (!(s = vid_match("", 1)))
				break;		// can't find any, give up
		check_abort();
		if (i++ > 50) {
			ati_maybe_blank(0);
			error("Can't switch from this video mode");
		}
		if (do_set_video(s)) {
			strcpy(current_video_mode, s);
			break;
		}
	} while (strcmp(name, s));	// give up if we get same mode twice
	ati_maybe_blank(0);
	note("Now %d x %d", screen_cols, screen_lines);
}

command set_video()
{
	char mode[30];

	if (is_gui)
		error("Resize a window by dragging its border.");
	else if (opsys == OS_UNIX || is_win32)
		error("Can't change video modes in this environment.");
	iter = 0;
	comp_read(mode, "Set video mode to: ", vid_match, MUST_MATCH, "");
	if (do_set_video(mode))
		strcpy(current_video_mode, mode);
	ati_maybe_blank(0);
	note("Now %d x %d", screen_cols, screen_lines);
}

resize_window_interactively(grow)
{
	int dir;

	for (;;) {
		say("Use arrow keys to %s this window, press %c to %s, "
			  "<Enter> when done.",
			  grow ? "enlarge" : "shrink", "+-"[grow],
			  grow ? "shrink" : "enlarge");
		getkey();
		check_abort();
		dir = get_direction();
		if (dir >= 0)
			alter_window_size(window_handle, dir, grow);
		if (binding_of(key, root_keys) == (short) help
			  || key == 'h' || key == '?')
			help_on_command("enlarge-window-interactively");
		switch (key) {
			case GREYPLUS:
			case '+':	grow = 1; break;
			case GREYMINUS:
			case '-':	grow = 0; break;
			case ' ':
			case GREYENTER:
			case '\n':
			case '\r':	say("");
						return;
		}
		maybe_refresh();
	}
}

int http_activity();

int url_operation(char *file, int op)
{
	struct url_parts parts;
	int port;
	char host_pat[FNAMELEN];

	prepare_url_operation(file, op, &parts);
	port = numtoi(parts.port);
	if (!strfcmp(parts.service, "http")) {
		char fname[FNAMELEN], auth[FNAMELEN];
		if (got_bad_number)
			port = 80;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTP files.");
		html_mode();
		when_net_activity = http_activity;
		*auth = 0;
		if (parts.have_usr)
			sprintf(auth, "%s:%s", parts.usr, parts.pwd);
		sprintf(host_pat, "|%s|", parts.host);
		if (http_proxy_port && !strstr(http_proxy_exceptions, parts.host)) {
			sprintf(fname, "http://%s:%d/%s", parts.host, port, parts.fname);
			http_retrieve(fname, http_proxy_server, http_proxy_port,
						  auth, bufname, 0);
		} else {
			sprintf(fname, "/%s", parts.fname);
			http_retrieve(fname, parts.host, port, auth, bufname, 0);
		}
	} else if (!strfcmp(parts.service, "telnet")) {
		if (got_bad_number)
			port = 23;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with telnet buffers.");
		do_telnet(parts.host, port, bufname);
	} else
		do_ftp_op(bufname, parts.host, parts.port, parts.usr, parts.pwd,
				  parts.fname, op);
	return 0;		// No error possible yet (asynchronous).
}

telnet_send_interrupt()
{
	telnet_send(telnet_id, "\03");
}

suffix_telnet()		// Needed for telnet:// url's to work.
{
	telnet_mode();
}

// The URL component s may have % escape sequences.  Translate them in
// place.

url_unescape(char *s)
{
	char *p, *hex = "0123456789abcdef", *d1, *d2;

	if (!s)
		return;
	while ((p = strchr(s, '%')) && p[1] && p[2]
		   && (d1 = strchr(hex, p[1])) && (d2 = strchr(hex, p[2]))) {
		*p = (d1 - hex) * 16 + (d2 - hex);
		strcpy(p + 1, p + 3);
	}
}

int parse_url(char *url, struct url_parts *p)
{
	memset(p, 0, sizeof(*p));
	strcpy(p->service, url);			// Parse service:\\ part.
	if (!split_string(p->service, ":", p->usr) || !*p->service
		  || is_path_separator(p->usr[0]) != 1
		  || is_path_separator(p->usr[1]) != 1)	// It's not valid.
		return 0;
	strcpy(p->usr, p->usr + 2);		// Split off the service part.
	split_string(p->usr, "/\\", p->fname);		// Remove the file name.

	if (!reverse_split_string(p->usr, "@", p->host)) {
										// Do we have a user part?
		strcpy(p->host, p->usr);		// No, so this must be the
		*p->usr = 0;					// host part.
	} else
		p->have_usr = 1;

	if (split_string(p->usr, ":", p->pwd))	// User part could have a pwd.
		p->have_password = 1;		// Pwd could be specified but empty.
	url_unescape(p->usr);
	url_unescape(p->pwd);
	if (split_string(p->host, ":", p->port))
		if (!*p->port)	// Host part could have a port (must be nonempty).
			return 0;
	return *p->host != 0;		// Host name must be nonempty.
}

color_tex_cmd(int s)
{
	char cmd[200];
	int color = color_class tex_literal;

	cmd[0] = '|';
	if (parse_string(1, "[a-z@]*", cmd + 1)) {
		strcat(cmd, "|");
		if (strstr("|begin|end|", cmd))
			color = color_class tex_begin_end;
		else if (strstr("|chapter|section|subsection|subsubsection"
						"|appendix|bibliography|title|tableofcontents"
						"|input|include|listoffigures|listoftables|part"
						"|usepackage|",
						cmd))
			color = color_class tex_section;
		else
			color = color_class tex_command;
	} else if (curchar() == '[') {
		int orig = point;
		if (re_search(1, "\n[ \t]*\n|\\par![^a-z]|\\]")
			  && character(point - 1) == ']') {
			color_tex_math(orig - 1, point, color_class tex_display_math);
			return;
		}
		point = orig + 1;
	} else
		point++;
	set_character_color(s, point, color);
}

color_tex_math(int from, int to, int color)
{
	int s;

	set_character_color(from, to, color);
	save_var narrow_start = from, narrow_end = size() - to, point = from;
	while (re_search(1, "[\\%&]")) {
		s = point - 1;		// Color % and & specially in math mode unless
		switch (character(s)) {	// protected by \ character.
			case '\\':
				point++;
				break;
			case '%':
				to_end_line();
				set_character_color(s, point, color_class tex_comment);
				break;
			case '&':
				set_character_color(s, point, color_class tex_section);
				break;
		}
	}
}

color_tex_dollar(int s)
{
	char pat[30];
	int orig, color = color_class tex_math;

	strcpy(pat, "\n[ \t]*\n|\\par![^a-z]|%$");
	if (curchar() == '$') {	// A $$ that signals displayed math?
		point++;			// Get to after $$.
		strcat(pat, "%$");
		color = color_class tex_display_math;
	}
	orig = point;
	if (re_search(1, pat) && character(point - 1) == '$')
		color_tex_math(s, point, color);
	else
		point = orig;
}

int fix_tex_math_coloring(int to)
{
	int c = get_character_color(point);

	if (c == color_class tex_math || c == color_class tex_display_math) {
		int s = point;			// Color rest of parag.
		if (!narrowed_search(REGEX, "\n[ \t]*\n", tex_look_back / 2))
			to_end_line();
		set_character_color(s, point, color_class tex_text);
		to = point;
		point = s;
	}
	return to;
}

tex_find_linked_file(char *file)
{
	char rel[FNAMELEN], curdir[FNAMELEN], *p, *include;

	to_begin_line();
	if (parse_string(1, ".*\\(include|input|usepackage)[ \t{]*([^{} \t\n]+)"))
		grab(find_group(2, 1), find_group(2, 0), rel);
	else
		error("Can't find any link here.");
	if (!*get_extension(rel))
		strcpy(get_extension(rel),
			   character(find_group(1, 1)) == 'u' ? ".sty" : ".tex");
	strcpy(file, rel);
	get_buffer_directory(curdir);
	absolute(file, curdir);
	if (check_file(file))
		return;
	include = getenv("TEXINPUTS");
	p = 0;
	if (!p && include)
		p = look_on_path(rel, 0, include);
	if (!p && *include_directories)
		p = look_on_path(rel, 0, include_directories);
	if (!p)
		error("%s not found in include-directories path", rel);
	strcpy(file, p);
}

int tex_maybe_break_line(int mode)
{
	switch (mode) {
		case 0:
		case 1:
			tex_auto_fill_mode = (mode != 0);
			return 0;
		case 2:
			return tex_auto_fill_mode;
	}
	if (!tex_auto_fill_mode)		// We're not in auto-fill mode.
		return 1;
	if (parse_string(-1, "([^\\]|<bob>)%%.*"))
		return 1;	// In comments, don't break lines.
	return maybe_break_this_line();
}

#include "inimode.e"
#include "makemode.e"
#include "misclang.e"
#include "gams.e"

// Load new default color settings.
color_class viewed_gui_text = color_class text;
color_class dialog_gui_text = color_class text;
color_class tex_display_math = color_class tex_math;

// For updating autoloaded files.
// If cmd isn't NULL, and names a command in the state file, then load
// the bytecode file (because its commands were already in the state file).
// Otherwise, autoload the bytecode file again.

update_autoload(char *cmd, char *file)
{
	char full[FNAMELEN], *s, olddir[FNAMELEN], newdir[FNAMELEN];

	strcpy(full, file);
	strcpy(get_extension(full), byte_extension);
	if (check_file(full))
		s = full;
	else
		s = lookpath(full);
	if (!s)
		error("Can't find %s", full);
	if (*cmd && name_type(find_index(cmd)) == NT_COMMAND)
		load_commands(s);
	else {
		getcd(olddir);
		strcpy(newdir, s);
		*get_tail(newdir, 0) = 0;
		chdir(newdir);
		autoload_commands(get_tail(s, 1));
		chdir(olddir);
	}
}

/*
Load this file to update your Epsilon state file to this version.
*/

when_loading()
{
	if (state_file && get_file_read_only(state_file) == 1) {
		save_var bufnum = zap(HELPBUF);
		if (strcmp(eel_version, EEL_MINOR_VERSION))
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"The above state file is for Epsilon version %s.\n"
					"This Epsilon executable is from Epsilon version %s.\n"
					"Perhaps EPSPATH is set incorrectly?\n",
					state_file, eel_version, EEL_MINOR_VERSION);
		else
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"But this Epsilon executable is from version %s,\n"
					"and so is the above state file.\n"
					"Therefore it appears no update is needed.\n",
					state_file, eel_version);
		view_help(" Update Not Performed ", 0);
		return;
	}

	// Load or autoload these as appropriate.
	update_autoload("info", "info");
	update_autoload("perl-mode", "perl");

	if (!strcmp(ignore_file_extensions, "|.obj|.exe|.b|.b2|.bu|"))
		strcpy(ignore_file_extensions, "|.obj|.exe|.o|.b|"); // new def
	drop_name("draw-column-marker");	// replaced by draw_column_markers
	reg_tab[ALT(DEL)] = find_index("backward_kill_word");

	strcpy(eel_version, EEL_MINOR_VERSION);

	delay(1, COND_KEY);	// Make sure the X window is up.
	refresh();
	sayput("Update complete, preparing to write state file...");
	delay(300, COND_KEY);	// Give the user a chance to abort.
	check_abort();
	if (!*state_file)
		write_state();		// Prompt for name.
	else
		do_save_state(state_file);	// Rewrite current state file.
	exit();
}
